"""
WebSocket - `RFC 6455 <https://www.rfc-editor.org/rfc/rfc6455>`_

Use the :func:`connect()` to establish a :class:`WebSocket` client connection.

Note from the developer: This is a very low-level API, which forces the
user to deal with things like data fragmentation.
A higher-level API could easily be built on top of this.

.. _authoring-callbacks:

Authoring Callbacks
-------------------
All network operations in `awscrt.websocket` are asynchronous.
Callbacks are always invoked on the WebSocket's networking thread.
You MUST NOT perform blocking network operations from any callback, or you will cause a deadlock.
For example: do not send a frame, and then wait for that frame to complete,
within a callback. The WebSocket cannot do work until your callback returns,
so the thread will be stuck. You can send the frame from within the callback,
just don't wait for it to complete within the callback.

If you want to do blocking waits, do it from a thread you control, like the main thread.
It's fine for the main thread to send a frame, and wait until it completes.

All functions and methods in `awscrt.websocket` are thread-safe.
They can be called from any mix of threads.

.. _flow-control-reading:

Flow Control (reading)
----------------------
By default, the WebSocket will read from the network as fast as it can hand you the data.
You must prevent the WebSocket from reading data faster than you can process it,
or memory usage could balloon until your application explodes.

There are two ways to manage this.

First, and simplest, is to process incoming data synchronously within the
`on_incoming_frame` callbacks. Since callbacks are invoked on the WebSocket's
networking thread, the WebSocket cannot read more data until the callback returns.
Therefore, processing the data in a synchronous manner
(i.e. writing to disk, printing to screen, etc) will naturally
affect `TCP flow control <https://en.wikipedia.org/wiki/Transmission_Control_Protocol#Flow_control>`_,
and prevent data from arriving too fast. However, you MUST NOT perform a blocking
network operation from within the callback or you risk deadlock (see :ref:`authoring-callbacks`).

The second, more complex, way requires you to manage the size of the read window.
Do this if you are processing the data asynchronously
(i.e. sending the data along on another network connection).
Create the WebSocket with `manage_read_window` set true,
and set `initial_read_window` to the number of bytes you are ready to receive right away.
Whenever the read window reaches 0, you will stop receiving anything.
The read window shrinks as you receive the payload from "data" frames (TEXT, BINARY, CONTINUATION).
Call :meth:`WebSocket.increment_read_window()` to increase the window again keep frames flowing in.
You only need to worry about the payload from "data" frames.
The WebSocket automatically increments its window to account for any
other incoming bytes, including other parts of a frame (opcode, payload-length, etc)
and the payload of other frame types (PING, PONG, CLOSE).
You'll probably want to do it like this:
Pick the max amount of memory to buffer, and set this as the `initial_read_window`.
When data arrives, the window has shrunk by that amount.
Send this data along on the other network connection.
When that data is done sending, call `increment_read_window()`
by the amount you just finished sending.
If you don't want to receive any data at first, set the `initial_read_window` to 0,
and `increment_read_window()` when you're ready.
Maintaining a larger window is better for overall throughput.

.. _flow-control-writing:

Flow Control (writing)
----------------------
You must also ensure that you do not continually send frames faster than the other
side can read them, or memory usage could balloon until your application explodes.

The simplest approach is to only send 1 frame at a time.
Use the :meth:`WebSocket.send_frame()` `on_complete` callback to know when the send is complete.
Then you can try and send another.

A more complex, but higher throughput, way is to let multiple frames be in flight
but have a cap. If the number of frames in flight, or bytes in flight, reaches
your cap then wait until some frames complete before trying to send more.

.. _api:

API
---
"""

# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0.
import _awscrt
from awscrt import NativeResource
import awscrt.exceptions
from awscrt.http import HttpProxyOptions, HttpRequest
from awscrt.io import ClientBootstrap, TlsConnectionOptions, SocketOptions
from dataclasses import dataclass
from enum import IntEnum
import sys
from typing import Callable, Optional, Sequence, Tuple, Union


class Opcode(IntEnum):
    """An opcode defines a frame's type.

    RFC 6455 classifies TEXT and BINARY as `data frames <https://www.rfc-editor.org/rfc/rfc6455#section-5.6>`_.
    A CONTINUATION frame "continues" the most recent data frame.
    All other opcodes are for `control frames <https://www.rfc-editor.org/rfc/rfc6455#section-5.5>`_.
    """

    CONTINUATION = 0x0,
    """Continues the most recent TEXT or BINARY data frame.

    See `RFC 6455 section 5.4 - Fragmentation <https://www.rfc-editor.org/rfc/rfc6455#section-5.4>`_.
    """

    TEXT = 0x1
    """The data frame for sending text.

    The payload must contain UTF-8."""

    BINARY = 0x2
    """The data frame for sending binary."""

    CLOSE = 0x8
    """The control frame which is the final frame sent by an endpoint.

    The CLOSE frame may include a payload, but its format is very particular.
    See `RFC 6455 section 5.5.1 <https://www.rfc-editor.org/rfc/rfc6455#section-5.5.1>`_.
    """

    PING = 0x9
    """A control frame that may serve as either a keepalive or as a means to verify
    that the remote endpoint is still responsive.

    DO NOT manually send a PONG frame in response to a PING,
    the implementation does this automatically.

    A PING frame may include a payload.
    See `RFC 6455 section 5.5.2 <https://www.rfc-editor.org/rfc/rfc6455#section-5.5.2>`_.
    """

    PONG = 0xA
    """The control frame that is the response to a PING frame.

    DO NOT manually send a PONG frame in response to a PING,
    the implementation does this automatically.

    See `RFC 6455 section 5.5.3 <https://www.rfc-editor.org/rfc/rfc6455#section-5.5.3>`_.
    """

    def is_data_frame(self):
        """True if this is a "data frame" opcode.

        TEXT, BINARY, and CONTINUATION are "data frames". The rest are "control" frames.

        If the WebSocket was created with `manage_read_window`,
        then the read window shrinks as "data frames" are received.
        See :ref:`flow-control-reading` for a thorough explanation.
        """
        return self.value in (Opcode.TEXT, Opcode.BINARY, Opcode.CONTINUATION)


MAX_PAYLOAD_LENGTH = 0x7FFFFFFFFFFFFFFF
"""The maximum frame payload length allowed by RFC 6455"""


@dataclass
class OnConnectionSetupData:
    """Data passed to the `on_connection_setup` callback"""

    exception: Optional[Exception] = None
    """If the connection failed, this exception explains why.

    This is None if the connection succeeded."""

    websocket: Optional['WebSocket'] = None
    """If the connection succeeded, here's the WebSocket.

    You should store this WebSocket somewhere
    (the connection will shut down if the class is garbage collected).

    This is None if the connection failed.
    """

    handshake_response_status: Optional[int] = None
    """The HTTP response status-code, if you're interested.

    This is present if an HTTP response was received, regardless of whether
    the handshake was accepted or rejected. This always has the value 101
    for successful connections.

    This is None if the connection failed before receiving an HTTP response.
    """

    handshake_response_headers: Optional[Sequence[Tuple[str, str]]] = None
    """The HTTP response headers, if you're interested.

    These are present if an HTTP response was received, regardless of whether
    the handshake was accepted or rejected.

    This is None if the connection failed before receiving an HTTP response.
    """

    handshake_response_body: bytes = None
    """The HTTP response body, if you're interested.

    This is only present if the server sent a full HTTP response rejecting the handshake.
    It is not present if the connection succeeded,
    or the connection failed for other reasons.
    """


@dataclass
class OnConnectionShutdownData:
    """Data passed to the `on_connection_shutdown` callback"""

    exception: Optional[Exception] = None
    """If the connection shut down cleanly, this is None.

    If the connection shut down due to error, or an error occurs while
    shutting down, this exception explains why."""


@dataclass
class IncomingFrame:
    """Describes the frame you are receiving.

    Used in `on_incoming_frame` callbacks """

    opcode: Opcode
    """This frame's opcode."""

    payload_length: int
    """This frame's payload length (in bytes)."""

    fin: bool
    """The FIN bit indicates whether this is the final fragment in a message.

    See `RFC 6455 section 5.4 - Fragmentation <https://www.rfc-editor.org/rfc/rfc6455#section-5.4>`_"""

    def is_data_frame(self):
        """True if this is a "data frame".

        TEXT, BINARY, and CONTINUATION are "data frames". The rest are "control frames".

        If the WebSocket was created with `manage_read_window`,
        then the read window shrinks as "data frames" are received.
        See :ref:`flow-control-reading` for a thorough explanation.
        """
        return self.opcode.is_data_frame()


@dataclass
class OnIncomingFrameBeginData:
    """Data passed to the `on_incoming_frame_begin` callback.

    Each `on_incoming_frame_begin` call will be followed by
    0+ `on_incoming_frame_payload` calls,
    followed by one `on_incoming_frame_complete` call."""

    frame: IncomingFrame
    """Describes the frame you are starting to receive."""


@dataclass
class OnIncomingFramePayloadData:
    """Data passed to the `on_incoming_frame_payload` callback.

    This callback will be invoked 0+ times.
    Each time, `data` will contain a bit more of the payload.
    Once all `frame.payload_length` bytes have been received
    (or the network connection is lost), the `on_incoming_frame_complete`
    callback will be invoked.

    If the WebSocket was created with `manage_read_window`,
    and this is a "data frame" (TEXT, BINARY, CONTINUATION),
    then the read window shrinks by `len(data)`.
    See :ref:`flow-control-reading` for a thorough explanation.
    """

    frame: IncomingFrame
    """Describes the frame whose payload you are receiving."""

    data: bytes
    """The next chunk of this frame's payload."""


@dataclass
class OnIncomingFrameCompleteData:
    """Data passed to the `on_incoming_frame_complete` callback."""

    frame: IncomingFrame
    """Describes the frame you are done receiving."""

    exception: Optional[Exception] = None
    """If `exception` is set, then something went wrong processing the frame
    or the connection was lost before the frame was fully received."""


@dataclass
class OnSendFrameCompleteData:
    """Data passed to the :meth:`WebSocket.send_frame()` `on_complete` callback."""

    exception: Optional[Exception] = None
    """If `exception` is set, the connection was lost before this frame could be completely sent.

    If `exception` is None, the frame was successfully written to the OS socket.
    Note that this data may still be buffered in the OS, it has
    not necessarily left this machine or reached the other endpoint yet."""


class WebSocket(NativeResource):
    """A WebSocket connection.

    Use :meth:`connect()` to establish a new client connection.
    """

    def __init__(self, binding):
        # Do not init a WebSocket directly, use websocket.connect()
        super().__init__()
        self._binding = binding

    def close(self):
        """Close the WebSocket asynchronously.

        You should call this when you are done with a healthy WebSocket,
        to ensure that it shuts down and cleans up.
        You don't need to call this on a WebSocket that has already shut
        down, or is in the middle of shutting down, but it is safe to do so.
        This function is idempotent.

        To determine when shutdown has completed, you can use the
        `on_shutdown_complete` callback (passed into :meth:`connect()`).
        """
        _awscrt.websocket_close(self._binding)

    def send_frame(
        self,
        opcode: Opcode,
        payload: Optional[Union[str, bytes, bytearray, memoryview]] = None,
        *,
        fin: bool = True,
        on_complete: Optional[Callable[[OnSendFrameCompleteData], None]] = None,
    ):
        """Send a WebSocket frame asynchronously.

        See `RFC 6455 section 5 - Data Framing <https://www.rfc-editor.org/rfc/rfc6455#section-5>`_
        for details on all frame types.

        This is a low-level API, which requires you to send the appropriate payload for each type of opcode.
        If you are not an expert, stick to sending :attr:`Opcode.TEXT` or :attr:`Opcode.BINARY` frames,
        and don't touch the FIN bit.

        See :ref:`flow-control-writing` to learn about limiting the amount of
        unsent data buffered in memory.

        Args:
            opcode: :class:`Opcode` for this frame.

            payload: Any `bytes-like object <https://docs.python.org/3/glossary.html#term-bytes-like-object>`_.
                `str` will always be encoded as UTF-8. It is fine to pass a `str` for a BINARY frame.
                None will result in an empty payload, the same as passing empty `bytes()`

            fin: The FIN bit indicates that this is the final fragment in a message.
                Do not set this False unless you understand
                `WebSocket fragmentation <https://www.rfc-editor.org/rfc/rfc6455#section-5.4>`_

            on_complete: Optional callback, invoked when the frame has finished sending.
                Takes a single :class:`OnSendFrameCompleteData` argument.

                If :attr:`OnSendFrameCompleteData.exception` is set, the connection
                was lost before this frame could be completely sent.

                But if `exception` is None, the frame was successfully written to the OS socket.
                (This doesn't mean the other endpoint has received the data yet,
                or even guarantee that the data has left the machine yet,
                but it's on track to get there).

                Be sure to read about :ref:`authoring-callbacks`.
        """
        def _on_complete(error_code):
            cbdata = OnSendFrameCompleteData()
            if error_code:
                cbdata.exception = awscrt.exceptions.from_code(error_code)

            # Do not let exceptions from the user's callback bubble up any further.
            try:
                if on_complete is not None:
                    on_complete(cbdata)
            except BaseException:
                print("Exception in WebSocket.send_frame on_complete callback", file=sys.stderr)
                sys.excepthook(*sys.exc_info())
                self.close()

        _awscrt.websocket_send_frame(
            self._binding,
            Opcode(opcode),  # needless cast to ensure opcode is valid
            payload,
            fin,
            _on_complete)

    def increment_read_window(self, size: int):
        """Manually increment the read window by this many bytes, to continue receiving frames.

        See :ref:`flow-control-reading` for a thorough explanation.
        If the WebSocket was created without `manage_read_window`, this function does nothing.
        This function may be called from any thread.

        Args:
            size: in bytes
        """
        if size < 0:
            raise ValueError("Increment size cannot be negative")

        _awscrt.websocket_increment_read_window(self._binding, size)


class _WebSocketCore(NativeResource):
    # Private class that handles wrangling callback data from C -> Python.
    # This class is kept alive by C until the final callback occurs.
    #
    # The only reason this class inherits from NativeResource,
    # is so our tests will tell us if the memory leaks.

    def __init__(self,
                 on_connection_setup,
                 on_connection_shutdown,
                 on_incoming_frame_begin,
                 on_incoming_frame_payload,
                 on_incoming_frame_complete):
        super().__init__()
        self._on_connection_setup_cb = on_connection_setup
        self._on_connection_shutdown_cb = on_connection_shutdown
        self._on_incoming_frame_begin_cb = on_incoming_frame_begin
        self._on_incoming_frame_payload_cb = on_incoming_frame_payload
        self._on_incoming_frame_complete_cb = on_incoming_frame_complete

    def _on_connection_setup(
            self,
            error_code,
            websocket_binding,
            handshake_response_status,
            handshake_response_headers,
            handshake_response_body):

        cbdata = OnConnectionSetupData()
        if error_code:
            cbdata.exception = awscrt.exceptions.from_code(error_code)
        else:
            cbdata.websocket = WebSocket(websocket_binding)

        cbdata.handshake_response_status = handshake_response_status
        cbdata.handshake_response_headers = handshake_response_headers
        cbdata.handshake_response_body = handshake_response_body

        # Do not let exceptions from the user's callback bubble up any further.
        try:
            self._on_connection_setup_cb(cbdata)
        except BaseException:
            print("Exception in WebSocket on_connection_setup callback", file=sys.stderr)
            sys.excepthook(*sys.exc_info())
            if cbdata.websocket is not None:
                cbdata.websocket.close()

    def _on_connection_shutdown(self, error_code):
        cbdata = OnConnectionShutdownData()
        if error_code:
            cbdata.exception = awscrt.exceptions.from_code(error_code)

        # Do not let exceptions from the user's callback bubble up any further.
        try:
            if self._on_connection_shutdown_cb is not None:
                self._on_connection_shutdown_cb(cbdata)
        except BaseException:
            print("Exception in WebSocket on_connection_shutdown callback", file=sys.stderr)
            sys.excepthook(*sys.exc_info())

    def _on_incoming_frame_begin(self, opcode_int, payload_length, fin):
        self._current_incoming_frame = IncomingFrame(Opcode(opcode_int), payload_length, fin)

        cbdata = OnIncomingFrameBeginData(self._current_incoming_frame)

        # Do not let exceptions from the user's callback bubble up any further:
        try:
            if self._on_incoming_frame_begin_cb is not None:
                self._on_incoming_frame_begin_cb(cbdata)
        except BaseException:
            print("Exception in WebSocket on_incoming_frame_begin callback", file=sys.stderr)
            sys.excepthook(*sys.exc_info())
            return False  # close websocket

        return True

    def _on_incoming_frame_payload(self, data):
        cbdata = OnIncomingFramePayloadData(self._current_incoming_frame, data)

        # Do not let exceptions from the user's callback bubble up any further:
        try:
            if self._on_incoming_frame_payload_cb is not None:
                self._on_incoming_frame_payload_cb(cbdata)
        except BaseException:
            print("Exception in WebSocket on_incoming_frame_payload callback", file=sys.stderr)
            sys.excepthook(*sys.exc_info())
            return False  # close websocket

        return True

    def _on_incoming_frame_complete(self, error_code):
        cbdata = OnIncomingFrameCompleteData(self._current_incoming_frame)
        if error_code:
            cbdata.exception = awscrt.exceptions.from_code(error_code)

        del self._current_incoming_frame

        # Do not let exceptions from the user's callback bubble up any further:
        try:
            if self._on_incoming_frame_complete_cb is not None:
                self._on_incoming_frame_complete_cb(cbdata)
        except BaseException:
            print("Exception in WebSocket on_incoming_frame_complete callback", file=sys.stderr)
            sys.excepthook(*sys.exc_info())
            return False  # close websocket

        return True


def connect(
    *,
    host: str,
    port: Optional[int] = None,
    handshake_request: HttpRequest,
    bootstrap: Optional[ClientBootstrap] = None,
    socket_options: Optional[SocketOptions] = None,
    tls_connection_options: Optional[TlsConnectionOptions] = None,
    proxy_options: Optional[HttpProxyOptions] = None,
    manage_read_window: bool = False,
    initial_read_window: Optional[int] = None,
    on_connection_setup: Callable[[OnConnectionSetupData], None],
    on_connection_shutdown: Optional[Callable[[OnConnectionShutdownData], None]] = None,
    on_incoming_frame_begin: Optional[Callable[[OnIncomingFrameBeginData], None]] = None,
    on_incoming_frame_payload: Optional[Callable[[OnIncomingFramePayloadData], None]] = None,
    on_incoming_frame_complete: Optional[Callable[[OnIncomingFrameCompleteData], None]] = None,
):
    """Asynchronously establish a client WebSocket connection.

    The `on_connection_setup` callback is invoked once the connection
    has succeeded or failed.

    If successful, a :class:`WebSocket` will be provided in the
    :class:`OnConnectionSetupData`. You should store this WebSocket somewhere,
    so that you can continue using it (the connection will shut down
    if the class is garbage collected).

    The WebSocket will shut down after one of these things occur:
        * You call :meth:`WebSocket.close()`
        * You, or the server, sends a CLOSE frame.
        * The underlying socket shuts down.
        * All references to the WebSocket are dropped,
          causing it to be garbage collected. However, you should NOT
          rely on this behavior. You should call :meth:`~WebSocket.close()` when you are
          done with a healthy WebSocket, to ensure that it shuts down and cleans up.
          It is very easy to accidentally keep a reference around without realizing it.

    Be sure to read about :ref:`authoring-callbacks`.

    Args:
        host: Hostname to connect to.

        port: Port to connect to. If not specified, it defaults to port 443
            when `tls_connection_options` is present, and port 80 otherwise.

        handshake_request: HTTP request for the initial WebSocket handshake.

            The request's method MUST be "GET", and the following headers are
            required::

                Host: <host>
                Upgrade: websocket
                Connection: Upgrade
                Sec-WebSocket-Key: <base64-encoding of 16 random bytes>
                Sec-WebSocket-Version: 13

            You can use :meth:`create_handshake_request()` to make a valid WebSocket
            handshake request, modifying the path and headers to fit your needs,
            and then passing it here.

        bootstrap: Client bootstrap to use when initiating socket connection.
            If not specified, the default singleton is used.

        socket_options: Socket options.
            If not specified, default options are used.

        proxy_options: HTTP Proxy options.
            If not specified, no proxy is used.

        manage_read_window: Set true to manually manage the flow-control read window.
            If false (the default), data arrives as fast as possible.
            See :ref:`flow-control-reading` for a thorough explanation.

        initial_read_window: The initial size of the read window, in bytes.
            This must be set if `manage_read_window` is true,
            otherwise it is ignored.
            See :ref:`flow-control-reading` for a thorough explanation.
            An initial size of 0 will prevent any frames from arriving
            until :meth:`WebSocket.increment_read_window()` is called.

        on_connection_setup: Callback invoked when the connect completes.
            Takes a single :class:`OnConnectionSetupData` argument.

            If successful, :attr:`OnConnectionSetupData.websocket` will be set.
            You should store the :class:`WebSocket` somewhere, so you can
            use it to send data when you're ready.
            The other callbacks will be invoked as events occur,
            until the final `on_connection_shutdown` callback.

            If unsuccessful, :attr:`OnConnectionSetupData.exception` will be set,
            and no further callbacks will be invoked.

            If this callback raises an exception, the connection will shut down.

        on_connection_shutdown: Optional callback, invoked when a connection shuts down.
            Takes a single :class:`OnConnectionShutdownData` argument.

            This callback is never invoked if `on_connection_setup` reported an exception.

        on_incoming_frame_begin: Optional callback, invoked once at the start of each incoming frame.
            Takes a single :class:`OnIncomingFrameBeginData` argument.

            Each `on_incoming_frame_begin` call will be followed by 0+
            `on_incoming_frame_payload` calls, followed by one
            `on_incoming_frame_complete` call.

            The "frame complete" callback is guaranteed to be invoked
            once for each "frame begin" callback, even if the connection
            is lost before the whole frame has been received.

            If this callback raises an exception, the connection will shut down.

        on_incoming_frame_payload: Optional callback, invoked 0+ times as payload data arrives.
            Takes a single :class:`OnIncomingFramePayloadData` argument.

            If `manage_read_window` is on, and this is a "data frame",
            then the read window shrinks accordingly.
            See :ref:`flow-control-reading` for a thorough explanation.

            If this callback raises an exception, the connection will shut down.

        on_incoming_frame_complete: Optional callback, invoked when the WebSocket
            is done processing an incoming frame.
            Takes a single :class:`OnIncomingFrameCompleteData` argument.

            If :attr:`OnIncomingFrameCompleteData.exception` is set,
            then something went wrong processing the frame
            or the connection was lost before the frame could be completed.

            If this callback raises an exception, the connection will shut down.
    """
    if manage_read_window:
        if initial_read_window is None:
            raise ValueError("'initial_read_window' must be set if 'manage_read_window' is enabled")
    else:
        initial_read_window = 0  # value is ignored anyway

    if initial_read_window < 0:
        raise ValueError("'initial_read_window' cannot be negative")

    if port is None:
        port = 0  # C layer uses zero to indicate "defaults please"

    if bootstrap is None:
        bootstrap = ClientBootstrap.get_or_create_static_default()

    if socket_options is None:
        socket_options = SocketOptions()

    core = _WebSocketCore(
        on_connection_setup,
        on_connection_shutdown,
        on_incoming_frame_begin,
        on_incoming_frame_payload,
        on_incoming_frame_complete)

    _awscrt.websocket_client_connect(
        host,
        port,
        handshake_request,
        bootstrap,
        socket_options,
        tls_connection_options,
        proxy_options,
        manage_read_window,
        initial_read_window,
        core)


def create_handshake_request(*, host: str, path: str = '/') -> HttpRequest:
    """Create an HTTP request with all the required fields for a WebSocket handshake.

    The method will be "GET", and the following headers are added::

        Host: <host>
        Upgrade: websocket
        Connection: Upgrade
        Sec-WebSocket-Key: <base64-encoding of 16 random bytes>
        Sec-WebSocket-Version: 13

    You may can add headers, or modify the path, before using this request.

    Args:
        host: Value for "Host" header
        path: Path (and query) string. Defaults to "/".
    """
    http_request_binding, http_headers_binding = _awscrt.websocket_create_handshake_request(host, path)
    return HttpRequest._from_bindings(http_request_binding, http_headers_binding)



"""
AWS client-side authentication: standard credentials providers and signing.
"""

# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0.

import _awscrt
from awscrt import NativeResource
import awscrt.exceptions
from awscrt.http import HttpRequest, HttpProxyOptions
from awscrt.io import ClientBootstrap, ClientTlsContext
from concurrent.futures import Future
import datetime
from enum import IntEnum
from typing import Optional, Sequence, Tuple


class AwsCredentials(NativeResource):
    """
    AwsCredentials are the public/private data needed to sign an authenticated AWS request.

    AwsCredentials are immutable.

    Args:
        access_key_id (str): Access key ID
        secret_access_key (str): Secret access key
        session_token (Optional[str]): Optional security token associated with
            the credentials.
        expiration (Optional[datetime.datetime]): Optional expiration datetime,
            that the credentials will no longer be valid past.
            Converted to UTC timezone and rounded down to nearest second.
            If not set, then credentials do not expire.

    Attributes:
        access_key_id (str): Access key ID
        secret_access_key (str): Secret access key
        session_token (Optional[str]): Security token associated with
            the credentials. None if not set.
        expiration (Optional[datetime.datetime]): Expiration datetime,
            that the credentials will no longer be valid past.
            None if credentials do not expire.
            Timezone is always UTC.
    """
    __slots__ = ()

    # C layer uses UINT64_MAX as timestamp for non-expiring credentials
    _NONEXPIRING_TIMESTAMP = 0xFFFFFFFFFFFFFFFF

    def __init__(self, access_key_id, secret_access_key, session_token=None, expiration=None):
        assert isinstance(access_key_id, str)
        assert isinstance(secret_access_key, str)
        assert isinstance(session_token, str) or session_token is None

        # C layer uses large int as timestamp for non-expiring credentials
        if expiration is None:
            expiration_timestamp = self._NONEXPIRING_TIMESTAMP
        else:
            expiration_timestamp = int(expiration.timestamp())
            if expiration_timestamp < 0 or expiration_timestamp >= self._NONEXPIRING_TIMESTAMP:
                raise OverflowError("expiration datetime out of range")

        super().__init__()
        self._binding = _awscrt.credentials_new(
            access_key_id,
            secret_access_key,
            session_token,
            expiration_timestamp)

    @classmethod
    def _from_binding(cls, binding):
        """Construct from a pre-existing native object"""
        credentials = cls.__new__(cls)  # avoid class's default constructor
        super(cls, credentials).__init__()  # just invoke parent class's __init__()
        credentials._binding = binding
        return credentials

    @property
    def access_key_id(self):
        return _awscrt.credentials_access_key_id(self._binding)

    @property
    def secret_access_key(self):
        return _awscrt.credentials_secret_access_key(self._binding)

    @property
    def session_token(self):
        return _awscrt.credentials_session_token(self._binding)

    @property
    def expiration(self):
        timestamp = _awscrt.credentials_expiration_timestamp_seconds(self._binding)
        # C layer uses large int as timestamp for non-expiring credentials
        if timestamp == self._NONEXPIRING_TIMESTAMP:
            return None
        else:
            return datetime.datetime.fromtimestamp(timestamp, tz=datetime.timezone.utc)

    def __deepcopy__(self, memo):
        # AwsCredentials is immutable, so just return self.
        return self


class AwsCredentialsProviderBase(NativeResource):
    # Pointless base class, kept for backwards compatibility.
    # AwsCredentialsProvider is (and always will be) the only subclass.
    #
    # Originally created with the thought that, when we supported
    # custom python providers, they would inherit from this class.
    # We ended up supporting custom python providers via
    # AwsCredentialsProvider.new_delegate() instead.
    pass


class AwsCredentialsProvider(AwsCredentialsProviderBase):
    """
    Credentials providers source the AwsCredentials needed to sign an authenticated AWS request.

    This class provides `new_X()` functions for several built-in provider types.
    To define a custom provider, use the :meth:`new_delegate()` function.
    """
    __slots__ = ()

    def __init__(self, binding):
        super().__init__()
        self._binding = binding

    @classmethod
    def new_default_chain(cls, client_bootstrap=None):
        """
        Create the default provider chain used by most AWS SDKs.

        Generally:

        1.  Environment
        2.  Profile
        3.  (conditional, off by default) ECS
        4.  (conditional, on by default) EC2 Instance Metadata

        Args:
            client_bootstrap (Optional[ClientBootstrap]): Client bootstrap to use when initiating socket connection.
                If not set, uses the default static ClientBootstrap instead.

        Returns:
            AwsCredentialsProvider:
        """
        assert isinstance(client_bootstrap, ClientBootstrap) or client_bootstrap is None

        if client_bootstrap is None:
            client_bootstrap = ClientBootstrap.get_or_create_static_default()

        binding = _awscrt.credentials_provider_new_chain_default(client_bootstrap)
        return cls(binding)

    @classmethod
    def new_static(cls, access_key_id, secret_access_key, session_token=None):
        """
        Create a simple provider that just returns a fixed set of credentials.

        Args:
            access_key_id (str): Access key ID
            secret_access_key (str): Secret access key
            session_token (Optional[str]): Optional session token

        Returns:
            AwsCredentialsProvider:
        """
        assert isinstance(access_key_id, str)
        assert isinstance(secret_access_key, str)
        assert isinstance(session_token, str) or session_token is None

        binding = _awscrt.credentials_provider_new_static(access_key_id, secret_access_key, session_token)
        return cls(binding)

    @classmethod
    def new_profile(
            cls,
            client_bootstrap=None,
            profile_name=None,
            config_filepath=None,
            credentials_filepath=None):
        """
        Creates a provider that sources credentials from key-value profiles
        loaded from the aws credentials file.

        Args:
            client_bootstrap (Optional[ClientBootstrap]): Client bootstrap to use when initiating socket connection.
                If not set, uses the static default ClientBootstrap instead.

            profile_name (Optional[str]): Name of profile to use.
                If not set, uses value from AWS_PROFILE environment variable.
                If that is not set, uses value of "default"

            config_filepath (Optional[str]): Path to profile config file.
                If not set, uses value from AWS_CONFIG_FILE environment variable.
                If that is not set, uses value of "~/.aws/config"

            credentials_filepath (Optional[str]): Path to profile credentials file.
                If not set, uses value from AWS_SHARED_CREDENTIALS_FILE environment variable.
                If that is not set, uses value of "~/.aws/credentials"

        Returns:
            AwsCredentialsProvider:
        """
        assert isinstance(client_bootstrap, ClientBootstrap) or client_bootstrap is None
        assert isinstance(profile_name, str) or profile_name is None
        assert isinstance(config_filepath, str) or config_filepath is None
        assert isinstance(credentials_filepath, str) or credentials_filepath is None

        if client_bootstrap is None:
            client_bootstrap = ClientBootstrap.get_or_create_static_default()

        binding = _awscrt.credentials_provider_new_profile(
            client_bootstrap, profile_name, config_filepath, credentials_filepath)
        return cls(binding)

    @classmethod
    def new_process(cls, profile_to_use=None):
        """
        Creates a provider that sources credentials from running an external command or process.

        The command to run is sourced from a profile in the AWS config file, using the standard
        profile selection rules. The profile key the command is read from is "credential_process."
        Example::

            [default]
            credential_process=/opt/amazon/bin/my-credential-fetcher --argsA=abc

        On successfully running the command, the output should be a json data with the following
        format::

            {
                "Version": 1,
                "AccessKeyId": "accesskey",
                "SecretAccessKey": "secretAccessKey"
                "SessionToken": "....",
                "Expiration": "2019-05-29T00:21:43Z"
            }

        Version here identifies the command output format version.
        This provider is not part of the default provider chain.

        Args:
            profile_to_use (Optional[str]): Name of profile in which to look for credential_process.
                If not set, uses value from AWS_PROFILE environment variable.
                If that is not set, uses value of "default"

        Returns:
            AwsCredentialsProvider:
        """

        binding = _awscrt.credentials_provider_new_process(profile_to_use)
        return cls(binding)

    @classmethod
    def new_environment(cls):
        """
        Creates a provider that returns credentials sourced from environment variables.

        * AWS_ACCESS_KEY_ID
        * AWS_SECRET_ACCESS_KEY
        * AWS_SESSION_TOKEN

        Returns:
            AwsCredentialsProvider:
        """

        binding = _awscrt.credentials_provider_new_environment()
        return cls(binding)

    @classmethod
    def new_chain(cls, providers):
        """
        Creates a provider that sources credentials from an ordered sequence of providers.

        This provider uses the first set of credentials successfully queried.
        Providers are queried one at a time; a provider is not queried until the
        preceding provider has failed to source credentials.

        Args:
            providers (List[AwsCredentialsProvider]): List of credentials providers.

        Returns:
            AwsCredentialsProvider:
        """

        binding = _awscrt.credentials_provider_new_chain(providers)
        return cls(binding)

    @classmethod
    def new_delegate(cls, get_credentials):
        """
        Creates a provider that sources credentials from a custom
        synchronous callback.

        Args:
            get_credentials: Callable which takes no arguments and returns
                :class:`AwsCredentials`.

        Returns:
            AwsCredentialsProvider:
        """
        # TODO: support async delegates

        assert callable(get_credentials)

        binding = _awscrt.credentials_provider_new_delegate(get_credentials)
        return cls(binding)

    @classmethod
    def new_cognito(
            cls,
            *,
            endpoint: str,
            identity: str,
            tls_ctx: awscrt.io.ClientTlsContext,
            logins: Optional[Sequence[Tuple[str, str]]] = None,
            custom_role_arn: Optional[str] = None,
            client_bootstrap: Optional[ClientBootstrap] = None,
            http_proxy_options: Optional[HttpProxyOptions] = None):
        """
        Creates a provider that sources credentials from the AWS Cognito Identity service.

        Args:
            endpoint (str): Cognito Identity service regional endpoint to source credentials from.

            identity (str): Cognito identity to fetch credentials relative to.

            tls_ctx (ClientTlsContext): Client TLS context to use when querying cognito credentials by HTTP.

            logins (Optional[Sequence[tuple[str, str]]]): Sequence of tuples specifying pairs of identity provider name
                and token values, representing established login contexts for identity authentication purposes.

            custom_role_arn (Optional[str]): ARN of the role to be assumed when multiple roles were received in the
                token from the identity provider.

            client_bootstrap (Optional[ClientBootstrap]): Client bootstrap to use when initiating a socket connection.
                If not set, uses the static default ClientBootstrap instead.

            http_proxy_options (Optional[HttpProxyOptions]): Optional HTTP proxy options.
                If None is provided then an HTTP proxy is not used.

        Returns:
            AwsCredentialsProvider:
        """

        assert isinstance(endpoint, str)
        assert isinstance(identity, str)
        assert isinstance(tls_ctx, ClientTlsContext)
        assert isinstance(custom_role_arn, str) or custom_role_arn is None
        assert isinstance(http_proxy_options, HttpProxyOptions) or http_proxy_options is None

        if client_bootstrap is None:
            client_bootstrap = ClientBootstrap.get_or_create_static_default()
        assert isinstance(client_bootstrap, ClientBootstrap)

        binding = _awscrt.credentials_provider_new_cognito(
            endpoint,
            identity,
            tls_ctx,
            client_bootstrap,
            logins,
            custom_role_arn,
            http_proxy_options)
        return cls(binding)

    @classmethod
    def new_x509(
            cls,
            *,
            endpoint: str,
            thing_name: str,
            role_alias: str,
            tls_ctx: awscrt.io.ClientTlsContext,
            client_bootstrap: Optional[ClientBootstrap] = None,
            http_proxy_options: Optional[HttpProxyOptions] = None):
        """
        Creates a provider that sources credentials from IoT's X509 credentials service.

        Args:
            endpoint (str): X509 service regional endpoint to source credentials from.
                            This is a per-account value that can be determined via the CLI:
                            `aws iot describe-endpoint --endpoint-type iot:CredentialProvider`

            thing_name (str): The name of the IoT thing to use to fetch credentials.

            role_alias (str): The name of the role alias to fetch credentials through.

            tls_ctx (ClientTlsContext): The client TLS context to use when establishing the http connection to IoT's X509 credentials service.

            client_bootstrap (Optional[ClientBootstrap]): Client bootstrap to use when initiating a socket connection.
                If not set, uses the static default ClientBootstrap instead.

            http_proxy_options (Optional[HttpProxyOptions]): Optional HTTP proxy options.
                If None is provided then an HTTP proxy is not used.

        Returns:
            AwsCredentialsProvider:
        """

        assert isinstance(endpoint, str)
        assert isinstance(thing_name, str)
        assert isinstance(role_alias, str)
        assert isinstance(tls_ctx, ClientTlsContext)
        assert isinstance(http_proxy_options, HttpProxyOptions) or http_proxy_options is None
        if client_bootstrap is None:
            client_bootstrap = ClientBootstrap.get_or_create_static_default()
        assert isinstance(client_bootstrap, ClientBootstrap)

        binding = _awscrt.credentials_provider_new_x509(
            endpoint,
            thing_name,
            role_alias,
            tls_ctx,
            client_bootstrap,
            http_proxy_options)
        return cls(binding)

    def get_credentials(self):
        """
        Asynchronously fetch AwsCredentials.

        Returns:
            concurrent.futures.Future: A Future which will contain
            :class:`AwsCredentials` (or an exception) when the operation completes.
            The operation may complete on a different thread.
        """
        future = Future()

        def _on_complete(error_code, binding):
            try:
                if error_code:
                    future.set_exception(awscrt.exceptions.from_code(error_code))
                else:
                    credentials = AwsCredentials._from_binding(binding)
                    future.set_result(credentials)

            except Exception as e:
                future.set_exception(e)

        try:
            _awscrt.credentials_provider_get_credentials(self._binding, _on_complete)
        except Exception as e:
            future.set_exception(e)

        return future


class AwsSigningAlgorithm(IntEnum):
    """AWS signing algorithm enumeration."""

    V4 = 0
    """Signature Version 4"""

    V4_ASYMMETRIC = 1
    """Signature Version 4 - Asymmetric"""

    V4_S3EXPRESS = 2
    """Signature Version 4 - S3 Express"""


class AwsSignatureType(IntEnum):
    """Which sort of signature should be computed from the signable."""

    HTTP_REQUEST_HEADERS = 0
    """
    A signature for a full HTTP request should be computed,
    with header updates applied to the signing result.
    """

    HTTP_REQUEST_QUERY_PARAMS = 1
    """
    A signature for a full HTTP request should be computed,
    with query param updates applied to the signing result.
    """


class AwsSignedBodyValue:
    """
    Values for use with :attr:`AwsSigningConfig.signed_body_value`.

    Some services use special values (e.g. "UNSIGNED-PAYLOAD") when the body
    is not being signed in the usual way.
    """

    EMPTY_SHA256 = 'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855'
    """The SHA-256 of the empty string."""

    UNSIGNED_PAYLOAD = 'UNSIGNED-PAYLOAD'
    """Unsigned payload option (not accepted by all services)"""

    STREAMING_AWS4_HMAC_SHA256_PAYLOAD = 'STREAMING-AWS4-HMAC-SHA256-PAYLOAD'
    """Each payload chunk will be signed (not accepted by all services)"""

    STREAMING_AWS4_HMAC_SHA256_EVENTS = 'STREAMING-AWS4-HMAC-SHA256-EVENTS'
    """Each event will be signed (not accepted by all services)"""


class AwsSignedBodyHeaderType(IntEnum):
    """
    Controls if signing adds a header containing the canonical request's signed body value.

    See :attr:`AwsSigningConfig.signed_body_value`.
    """

    NONE = 0
    """Do not add a header."""

    X_AMZ_CONTENT_SHA_256 = 1
    """Add the "x-amz-content-sha256" header with the canonical request's signed body value"""


class AwsSigningConfig(NativeResource):
    """
    Configuration for use in AWS-related signing.

    AwsSigningConfig is immutable.

    It is good practice to use a new config for each signature, or the date might get too old.

    Args:
        algorithm (AwsSigningAlgorithm): Which signing algorithm to use.

        signature_type (AwsSignatureType): Which sort of signature should be
            computed from the signable.

        credentials_provider (AwsCredentialsProvider): Credentials provider
            to fetch signing credentials with. If the algorithm is
            :attr:`AwsSigningAlgorithm.V4_ASYMMETRIC`, ECC-based credentials will be derived from the
            fetched credentials.

        region (str): If the algorithm is :attr:`AwsSigningAlgorithm.V4`, the region to sign against.
            If the algorithm is :attr:`AwsSigningAlgorithm.V4_ASYMMETRIC`, the value of the
            "X-amzn-region-set" header (added in signing).

        service (str): Name of service to sign a request for.

        date (Optional[datetime.datetime]): Date and time to use during the
            signing process. If None is provided then
            `datetime.datetime.now(datetime.timezone.utc)` is used.
            Naive dates (lacking timezone info) are assumed to be in local time.

        should_sign_header (Optional[Callable[[str], bool]]):
            Optional function to control which headers are
            a part of the canonical request.

            Skipping auth-required headers will result in an unusable signature.
            Headers injected by the signing process are not skippable.
            This function does not override the internal check function
            (x-amzn-trace-id, user-agent), but rather supplements it.
            In particular, a header will get signed if and only if it returns
            true to both the internal check (skips x-amzn-trace-id, user-agent)
            and this function (if defined).

        use_double_uri_encode (bool): Whether to double-encode the resource path
            when constructing the canonical request (assuming the path is already
            encoded). Default is True. All services except S3 use double encoding.

        should_normalize_uri_path (bool): Whether the resource paths are
            normalized when building the canonical request. Default is True.

        signed_body_value (Optional[str]): If set, this value is used as the
            canonical request's body value. Typically, this is the SHA-256
            of the payload, written as lowercase hex. If this has been
            precalculated, it can be set here. Special values used by certain
            services can also be set (see :class:`AwsSignedBodyValue`). If `None`
            is passed (the default), the typical value will be calculated from
            the payload during signing.

        signed_body_header_type (AwsSignedBodyHeaderType): Controls if signing
            adds a header containing the canonical request's signed body value.
            Default is to not add a header.

        expiration_in_seconds (Optional[int]): If set, and signature_type is
            :attr:`AwsSignatureType.HTTP_REQUEST_QUERY_PARAMS`, then signing will add "X-Amz-Expires"
            to the query string, equal to the value specified here.

        omit_session_token (bool): If set True, the "X-Amz-Security-Token"
            query param is omitted from the canonical request.
            The default False should be used for most services.
    """
    __slots__ = ('_priv_should_sign_cb')

    _attributes = (
        'algorithm',
        'signature_type',
        'credentials_provider',
        'region',
        'service',
        'date',
        'should_sign_header',
        'use_double_uri_encode',
        'should_normalize_uri_path',
        'signed_body_value',
        'signed_body_header_type',
        'expiration_in_seconds',
        'omit_session_token',
    )

    def __init__(self,
                 algorithm=AwsSigningAlgorithm.V4,
                 signature_type=AwsSignatureType.HTTP_REQUEST_HEADERS,
                 credentials_provider=None,
                 region="",
                 service="",
                 date=None,
                 should_sign_header=None,
                 use_double_uri_encode=True,
                 should_normalize_uri_path=True,
                 signed_body_value=None,
                 signed_body_header_type=AwsSignedBodyHeaderType.NONE,
                 expiration_in_seconds=None,
                 omit_session_token=False,
                 ):

        assert isinstance(algorithm, AwsSigningAlgorithm)
        assert isinstance(signature_type, AwsSignatureType)
        assert isinstance(credentials_provider, AwsCredentialsProvider) or credentials_provider is None
        assert isinstance(region, str)
        assert isinstance(service, str)
        assert callable(should_sign_header) or should_sign_header is None
        assert signed_body_value is None or (isinstance(signed_body_value, str) and len(signed_body_value) > 0)
        assert isinstance(signed_body_header_type, AwsSignedBodyHeaderType)
        assert expiration_in_seconds is None or expiration_in_seconds > 0

        super().__init__()

        if date is None:
            date = datetime.datetime.now(datetime.timezone.utc)

        timestamp = date.timestamp()

        self._priv_should_sign_cb = should_sign_header

        if should_sign_header is not None:
            def should_sign_header_wrapper(name):
                return should_sign_header(name=name)
        else:
            should_sign_header_wrapper = None

        if expiration_in_seconds is None:
            # C layer uses 0 to indicate None
            expiration_in_seconds = 0

        self._binding = _awscrt.signing_config_new(
            algorithm,
            signature_type,
            credentials_provider,
            region,
            service,
            date,
            timestamp,
            should_sign_header_wrapper,
            use_double_uri_encode,
            should_normalize_uri_path,
            signed_body_value,
            signed_body_header_type,
            expiration_in_seconds,
            omit_session_token)

    def replace(self, **kwargs):
        """
        Return an AwsSigningConfig with the same attributes, except for those
        attributes given new values by whichever keyword arguments are specified.
        """
        args = {x: kwargs.get(x, getattr(self, x)) for x in AwsSigningConfig._attributes}
        return AwsSigningConfig(**args)

    @property
    def algorithm(self):
        """AwsSigningAlgorithm: Which signing algorithm to use"""
        return AwsSigningAlgorithm(_awscrt.signing_config_get_algorithm(self._binding))

    @property
    def signature_type(self):
        """AwsSignatureType: Which sort of signature should be computed from the signable."""
        return AwsSignatureType(_awscrt.signing_config_get_signature_type(self._binding))

    @property
    def credentials_provider(self):
        """
        AwsCredentialsProvider: Credentials provider to fetch signing credentials with.
        If the algorithm is :attr:`AwsSigningAlgorithm.V4_ASYMMETRIC`, ECC-based credentials will be derived
        from the fetched credentials.
        """
        return _awscrt.signing_config_get_credentials_provider(self._binding)

    @property
    def region(self):
        """
        str: If signing algorithm is :attr:`AwsSigningAlgorithm.V4`, the region to sign against.
        If the algorithm is :attr:`AwsSigningAlgorithm.V4_ASYMMETRIC`, the value of the
        "X-amzn-region-set header" (added in signing).
        """
        return _awscrt.signing_config_get_region(self._binding)

    @property
    def service(self):
        """str: Name of service to sign a request for"""
        return _awscrt.signing_config_get_service(self._binding)

    @property
    def date(self):
        """
        datetime.datetime: Date and time to use during the signing process.

        If None is provided, then `datetime.datetime.now(datetime.timezone.utc)`
        at time of object construction is used.

        It is good practice to use a new config for each signature, or the date might get too old.
        """
        return _awscrt.signing_config_get_date(self._binding)

    @property
    def should_sign_header(self):
        """
        Optional[Callable[[str], bool]]: Optional function to control which
        headers are a part of the canonical request.

        Skipping auth-required headers will result in an unusable signature.
        Headers injected by the signing process are not skippable.
        This function does not override the internal check function
        (x-amzn-trace-id, user-agent), but rather supplements it. In particular,
        a header will get signed if and only if it returns true to both
        the internal check (skips x-amzn-trace-id, user-agent) and this function (if defined).
        """
        return self._priv_should_sign_cb

    @property
    def use_double_uri_encode(self):
        """
        bool: Whether to double-encode the resource path when constructing
        the canonical request (assuming the path is already encoded).

        By default, all services except S3 use double encoding.
        """
        return _awscrt.signing_config_get_use_double_uri_encode(self._binding)

    @property
    def should_normalize_uri_path(self):
        """
        bool: Whether the resource paths are normalized when building the
        canonical request.
        """
        return _awscrt.signing_config_get_should_normalize_uri_path(self._binding)

    @property
    def signed_body_value(self):
        """
        Optional[str]: What to use as the canonical request's body value.
        If `None` is set (the default), a value will be calculated from
        the payload during signing. Typically, this is the SHA-256 of the
        payload, written as lowercase hex. If this has been precalculated,
        it can be set here. Special values used by certain services can also
        be set (see :class:`AwsSignedBodyValue`).
        """
        return _awscrt.signing_config_get_signed_body_value(self._binding)

    @property
    def signed_body_header_type(self):
        """
        AwsSignedBodyHeaderType: Controls if signing adds a header containing
        the canonical request's signed body value.
        """
        return AwsSignedBodyHeaderType(_awscrt.signing_config_get_signed_body_header_type(self._binding))

    @property
    def expiration_in_seconds(self):
        """
        Optional[int]: If set, and signature_type is :attr:`AwsSignatureType.HTTP_REQUEST_QUERY_PARAMS`,
        then signing will add "X-Amz-Expires" to the query string, equal to the
        value specified here. Otherwise, this is None has no effect.
        """
        expiration = _awscrt.signing_config_get_expiration_in_seconds(self._binding)
        # C layer uses 0 to indicate None
        return None if expiration == 0 else expiration

    @property
    def omit_session_token(self):
        """
        bool: Whether the "X-Amz-Security-Token" query param is omitted
        from the canonical request. This should be False for most services.
        """
        return _awscrt.signing_config_get_omit_session_token(self._binding)


def aws_sign_request(http_request, signing_config):
    """
    Perform AWS HTTP request signing.

    The :class:`awscrt.http.HttpRequest` is transformed asynchronously,
    according to the :class:`AwsSigningConfig`.

    When signing:

    1.  It is good practice to use a new config for each signature,
        or the date might get too old.

    2.  Do not add the following headers to requests before signing, they may be added by the signer:
        x-amz-content-sha256,
        X-Amz-Date,
        Authorization

    3.  Do not add the following query params to requests before signing, they may be added by the signer:
        X-Amz-Signature,
        X-Amz-Date,
        X-Amz-Credential,
        X-Amz-Algorithm,
        X-Amz-SignedHeaders

    Args:
        http_request (awscrt.http.HttpRequest): The HTTP request to sign.
        signing_config (AwsSigningConfig): Configuration for signing.

    Returns:
        concurrent.futures.Future: A Future whose result will be the signed
        :class:`awscrt.http.HttpRequest`. The future will contain an exception
        if the signing process fails.
    """

    assert isinstance(http_request, HttpRequest)
    assert isinstance(signing_config, AwsSigningConfig)

    future = Future()

    def _on_complete(error_code):
        try:
            if error_code:
                future.set_exception(awscrt.exceptions.from_code(error_code))
            else:
                future.set_result(http_request)
        except Exception as e:
            future.set_exception(e)

    _awscrt.sign_request_aws(http_request, signing_config, _on_complete)
    return future



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0.

"""Python logging integration for the AWS CRT."""

import logging
import threading
from enum import IntEnum

import _awscrt
from awscrt.io import LogLevel

__all__ = ['init_logging', 'set_log_level', 'logf', 'LogSubject', 'CRT_LOG_FORMAT']

_CRT_TO_PY_LEVEL = {
    0: logging.NOTSET,
    1: logging.CRITICAL,
    2: logging.ERROR,
    3: logging.WARNING,
    4: logging.INFO,
    5: logging.DEBUG,
    6: logging.DEBUG,
}

_PY_TO_CRT_LEVEL = {
    logging.NOTSET: LogLevel.NoLogs,
    logging.CRITICAL: LogLevel.Fatal,
    logging.ERROR: LogLevel.Error,
    logging.WARNING: LogLevel.Warn,
    logging.INFO: LogLevel.Info,
    logging.DEBUG: LogLevel.Trace,
}

# Default Format supported by native loggers.
CRT_LOG_FORMAT = '%(asctime)s [%(threadName)s] %(levelname)s %(name)s - %(message)s'


class LogSubject(IntEnum):
    """Log subject identifiers for CRT subsystems."""
    # aws-c-common (package ID 0)
    COMMON_GENERAL = 0x000
    COMMON_TASK_SCHEDULER = 0x001

    # aws-c-io (package ID 1)
    IO_GENERAL = 0x400
    IO_EVENT_LOOP = 0x401
    IO_SOCKET = 0x402
    IO_SOCKET_HANDLER = 0x403
    IO_TLS = 0x404
    IO_ALPN = 0x405
    IO_DNS = 0x406
    IO_PKI = 0x407
    IO_CHANNEL = 0x408
    IO_CHANNEL_BOOTSTRAP = 0x409
    IO_FILE_UTILS = 0x40A
    IO_SHARED_LIBRARY = 0x40B

    # aws-c-http (package ID 2)
    HTTP_GENERAL = 0x800
    HTTP_CONNECTION = 0x801
    HTTP_SERVER = 0x802
    HTTP_STREAM = 0x803
    HTTP_CONNECTION_MANAGER = 0x804
    HTTP_WEBSOCKET = 0x805
    HTTP_WEBSOCKET_SETUP = 0x806

    # aws-c-event-stream (package ID 4)
    EVENT_STREAM_GENERAL = 0x1000
    EVENT_STREAM_CHANNEL_HANDLER = 0x1001
    EVENT_STREAM_RPC_SERVER = 0x1002
    EVENT_STREAM_RPC_CLIENT = 0x1003

    # aws-c-mqtt (package ID 5)
    MQTT_GENERAL = 0x1400
    MQTT_CLIENT = 0x1401
    MQTT_TOPIC_TREE = 0x1402

    # aws-c-auth (package ID 6)
    AUTH_GENERAL = 0x1800
    AUTH_PROFILE = 0x1801
    AUTH_CREDENTIALS_PROVIDER = 0x1802
    AUTH_SIGNING = 0x1803

    # aws-c-cal (package ID 7)
    CAL_GENERAL = 0x1C00
    CAL_ECC = 0x1C01
    CAL_HASH = 0x1C02
    CAL_HMAC = 0x1C03
    CAL_DER = 0x1C04
    CAL_LIBCRYPTO_RESOLVE = 0x1C05
    CAL_RSA = 0x1C06
    CAL_ED25519 = 0x1C07

    # aws-c-s3 (package ID 14)
    S3_GENERAL = 0x3800
    S3_CLIENT = 0x3801

    # aws-c-sdkutils (package ID 15)
    SDKUTILS_GENERAL = 0x3C00
    SDKUTILS_PROFILE = 0x3C01
    SDKUTILS_ENDPOINTS_PARSING = 0x3C02
    SDKUTILS_ENDPOINTS_RESOLVE = 0x3C03
    SDKUTILS_ENDPOINTS_GENERAL = 0x3C04
    SDKUTILS_PARTITIONS_PARSING = 0x3C05
    SDKUTILS_ENDPOINTS_REGEX = 0x3C06


_PACKAGE_ID_TO_MODULE = {
    0: 'common',
    1: 'io',
    2: 'http',
    4: 'event-stream',
    5: 'mqtt',
    6: 'auth',
    7: 'cal',
    14: 's3',
    15: 'sdkutils',
}


def _python_logging_callback(crt_level, message, subject_id, subject_name, thread_name):
    """Called from C for each CRT log message."""
    module = _PACKAGE_ID_TO_MODULE.get(subject_id >> 10, 'unknown')
    logger = logging.getLogger('awscrt.{}.{}'.format(module, subject_name))
    py_level = _CRT_TO_PY_LEVEL.get(crt_level, logging.DEBUG)
    # `There is the possibility that “dummy thread objects” are created. These are
    # thread objects corresponding to “alien threads”, which are threads of control
    # started outside the threading module, such as directly from C code.`
    # https://docs.python.org/3/library/threading.html#thread-objects
    #
    # As per current conventions, dummy thread objects have thread names
    # starting with "Dummy". Eg. Dummy-1, Dummy-8
    if threading.current_thread().name.startswith("Dummy") and thread_name is not None:
        threading.current_thread().name = thread_name
    logger.log(py_level, '%s', message)


def init_logging(level: int = logging.DEBUG):
    """Initialize CRT logging, routing output through Python's logging module.

    Log messages appear under the ``awscrt`` logger hierarchy, with each CRT
    subsystem as a child logger (e.g. ``awscrt.io.event-loop``,
    ``awscrt.s3.S3Client``).

    This is mutually exclusive with :func:`~awscrt.io.init_logging` -- use one
    or the other, not both. Can only be called once.

    Example usage::

        import logging
        from awscrt.logging import init_logging

        logging.basicConfig(level=logging.DEBUG)
        init_logging(logging.DEBUG)

    Args:
        level (int): Python logging level (e.g. ``logging.DEBUG``,
            ``logging.WARNING``).

    Raises:
        RuntimeError: If CRT logging has already been initialized.
    """
    root_logger = logging.getLogger('awscrt')

    try:
        crt_level = _PY_TO_CRT_LEVEL[level]
    except KeyError:
        raise ValueError(f"Invalid log level: {level}. Use logging.DEBUG, INFO, WARNING, ERROR, or CRITICAL")

    if root_logger.level == logging.NOTSET:
        root_logger.setLevel(_CRT_TO_PY_LEVEL.get(int(crt_level), logging.DEBUG))

    _awscrt.init_python_logging(int(crt_level), _python_logging_callback)


def set_log_level(level: int):
    """Change the CRT log level. :func:`init_logging` must have been called first.

    Set log level to logging.NOTSET to disable the logger. Cleaning up the logger
    is dangerous in a multi-threaded environment.

    Args:
        level (int): Python logging level (e.g. ``logging.DEBUG``,
            ``logging.WARNING``).
    """
    try:
        crt_level = _PY_TO_CRT_LEVEL[level]
    except KeyError:
        raise ValueError(f"Invalid log level: {level}. Use logging.DEBUG, INFO, WARNING, ERROR, or CRITICAL")
    _awscrt.set_log_level(int(crt_level))


def log(level: int, subject: LogSubject, message: str):
    """Log a message through the CRT's configured logger.

    Args:
        level (int): Python logging level (e.g. logging.DEBUG, logging.INFO).
        subject (LogSubject): Log subject identifying the subsystem.
        message (str): The message to log.
    """
    try:
        crt_level = _PY_TO_CRT_LEVEL[level]
    except KeyError:
        raise ValueError(f"Invalid log level: {level}. Use logging.DEBUG, INFO, WARNING, ERROR, or CRITICAL")
    _awscrt.logger_log(int(crt_level), int(subject), message)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0.

from dataclasses import dataclass


@dataclass
class AWSIoTMetrics:
    """
    Configuration for IoT SDK metrics that are embedded in MQTT Connect Packet username field.

    Args:
        library_name (str): The SDK library name (e.g., "IoTDeviceSDK/Python")

    """
    library_name: str = "IoTDeviceSDK/Python"
